#include <cstdio>
#include <cassert>
#include "Brain.hpp"
#include "File.hpp"

Instruction::Instruction()
: m_instructionType(InstructionType_Sense)
{
}

Instruction::~Instruction()
{
}

int Instruction::getNumArgs() const
{
  return m_args.size();
}

void Instruction::setNumArgs(int numArgs)
{
  m_args.resize(numArgs);
}

int Instruction::getArg(int argNum) const
{
  assert(argNum >= 0 && argNum < (int)m_args.size());
  return m_args[argNum];
}

void Instruction::setArg(int argNum, int arg)
{
  assert(argNum >= 0 && argNum < (int)m_args.size());
  m_args[argNum] = arg;
}

Brain::Brain()
{
}

Brain::~Brain()
{
}

int Brain::getNumInstructions() const
{
  return m_instructions.size();
}

const Instruction &Brain::getInstruction(int instructionNum) const
{
  assert(instructionNum >= 0 && instructionNum < (int)m_instructions.size());
  return m_instructions[instructionNum];
}

void Brain::addInstruction(const Instruction &instruction)
{
  m_instructions.push_back(instruction);
}

namespace
{
  void extractLine(const vector<unsigned char> buf,
                   int *pPos,
                   vector<string> *pTokens)
  {
    int &pos = *pPos;
    int sz = buf.size();

    bool newString = true;
    bool comment = false;

    while (pos < sz && buf[pos] != '\n') {
      unsigned char ch = buf[pos];
      pos++;

      if (!comment) {
        if (ch == ';') {
          comment = true;
        } else if (ch == ' ' || ch == '\t') {
          newString = true;
        } else {
          if (newString) {
            pTokens->resize(pTokens->size() + 1);
            newString = false;
          }
          (*pTokens)[pTokens->size() - 1].push_back(ch);
        }
      }
    }

    if (pos < sz) {
      pos++;
    }
  }

  bool equal(const string &str1, const string &str2)
  {
    int sz = (int)str1.size();
    if (sz != (int)str2.size()) {
      return false;
    }
    for (int i = 0; i < sz; i++) {
      char ch1 = str1[i];
      char ch2 = str2[i];
      if (ch1 == ch2 ||
          (ch1 >= 'A' && ch1 <= 'Z' && ch1 - 'A' + 'a' == ch2) ||
          (ch1 >= 'a' && ch1 <= 'z' && ch1 - 'a' + 'A' == ch2)) {
      } else {
        return false;
      }
    }
    return true;
  }

  bool parseInstructionType(const string &token,
                            InstructionType *pInstructionType)
  {
    if (equal(token, "Sense")) {
      *pInstructionType = InstructionType_Sense;
    } else if (equal(token, "Mark")) {
      *pInstructionType = InstructionType_Mark;
    } else if (equal(token, "Unmark")) {
      *pInstructionType = InstructionType_Unmark;
    } else if (equal(token, "PickUp")) {
      *pInstructionType = InstructionType_PickUp;
    } else if (equal(token, "Drop")) {
      *pInstructionType = InstructionType_Drop;
    } else if (equal(token, "Turn")) {
      *pInstructionType = InstructionType_Turn;
    } else if (equal(token, "Move")) {
      *pInstructionType = InstructionType_Move;
    } else if (equal(token, "Flip")) {
      *pInstructionType = InstructionType_Flip;
    } else {
      return false;
    }
    return true;
  }

  bool parseSenseDir(const string &token,
                     SenseDir *pSenseDir)
  {
    if (equal(token, "Here")) {
      *pSenseDir = SenseDir_Here;
    } else if (equal(token, "Ahead")) {
      *pSenseDir = SenseDir_Ahead;
    } else if (equal(token, "LeftAhead")) {
      *pSenseDir = SenseDir_LeftAhead;
    } else if (equal(token, "RightAhead")) {
      *pSenseDir = SenseDir_RightAhead;
    } else {
      return false;
    }
    return true;
  }

  bool formatSenseDir(SenseDir senseDir,
                      string *pStr) {
    switch (senseDir) {
    case SenseDir_Here:
      *pStr = "Here"; return true;
    case SenseDir_Ahead:
      *pStr = "Ahead"; return true;
    case SenseDir_LeftAhead:
      *pStr = "LeftAhead"; return true;
    case SenseDir_RightAhead:
      *pStr = "RightAhead"; return true;
    default:
      return false;
    }
  }

  bool parseCond(const string &token,
                 Cond *pCond)
  {
    if (equal(token, "Friend")) {
      *pCond = Cond_Friend;
    } else if (equal(token, "Foe")) {
      *pCond = Cond_Foe;
    } else if (equal(token, "FriendWithFood")) {
      *pCond = Cond_FriendWithFood;
    } else if (equal(token, "FoeWithFood")) {
      *pCond = Cond_FoeWithFood;
    } else if (equal(token, "Food")) {
      *pCond = Cond_Food;
    } else if (equal(token, "Rock")) {
      *pCond = Cond_Rock;
    } else if (equal(token, "Marker")) {
      *pCond = Cond_Marker;
    } else if (equal(token, "FoeMarker")) {
      *pCond = Cond_FoeMarker;
    } else if (equal(token, "Home")) {
      *pCond = Cond_Home;
    } else if (equal(token, "FoeHome")) {
      *pCond = Cond_FoeHome;
    } else {
      return false;
    }
    return true;
  }

  bool formatCond(Cond cond,
                  string *pStr) {
    switch (cond) {
    case Cond_Friend:
      *pStr = "Friend"; return true;
    case Cond_Foe:
      *pStr = "Foe"; return true;
    case Cond_FriendWithFood:
      *pStr = "FriendWithFood"; return true;
    case Cond_FoeWithFood:
      *pStr = "FoeWithFood"; return true;
    case Cond_Food:
      *pStr = "Food"; return true;
    case Cond_Rock:
      *pStr = "Rock"; return true;
    case Cond_Marker:
      *pStr = "Marker"; return true;
    case Cond_FoeMarker:
      *pStr = "FoeMarker"; return true;
    case Cond_Home:
      *pStr = "Home"; return true;
    case Cond_FoeHome:
      *pStr = "FoeHome"; return true;
    default:
      return false;
    }
  }

  bool parseLR(const string &token,
               LR *pLR)
  {
    if (equal(token, "Left")) {
      *pLR = LR_Left;
    } else if (equal(token, "Right")) {
      *pLR = LR_Right;
    } else {
      return false;
    }
    return true;
  }

  bool formatLR(LR lr,
                string *pStr) {
    switch (lr) {
    case LR_Left:
      *pStr = "Left"; return true;
    case LR_Right:
      *pStr = "Right"; return true;
    default:
      return false;
    }
  }

  bool parseInt(const string &token,
                int *pInt)
  {
    int sz = token.size();
    int result = 0;
    for (int i = 0; i < sz; i++) {
      char ch = token[i];
      if (ch < '0' || ch > '9') {
        return false;
      }
      result *= 10;
      result += (ch - '0');
    }
    *pInt = result;
    return true;
  }

  bool formatInt(int i,
                 string *pStr)
  {
    char buf[1000];
    sprintf(buf, "%d", i);
    *pStr = buf;
    return true;
  }
}

bool readBrain(const string &fileName, Brain *pBrain)
{
  vector<unsigned char> buf;
  if (!readFile(fileName, &buf)) {
    return false;
  }

  int pos = 0;
  int sz = buf.size();

  while (pos < sz) {
    vector<string> tokens;
    extractLine(buf, &pos, &tokens);

    if (tokens.size() < 1) {
      return false;
    }

    Instruction instruction;

    InstructionType instructionType;
    if (!parseInstructionType(tokens[0], &instructionType)) {
      return false;
    }
    instruction.setInstructionType(instructionType);

    vector<string> args(tokens);
    args.erase(args.begin());
    instruction.setNumArgs(args.size());

    switch (instructionType) {
    case InstructionType_Sense:
      {
        if (args.size() != 4 && args.size() != 5) {
          return false;
        }

        SenseDir senseDir;
        if (!parseSenseDir(args[0], &senseDir)) {
          return false;
        }
        instruction.setArg(0, senseDir);

        int st1;
        if (!parseInt(args[1], &st1)) {
          return false;
        }
        instruction.setArg(1, st1);

        int st2;
        if (!parseInt(args[2], &st2)) {
          return false;
        }
        instruction.setArg(2, st2);

        Cond cond;
        if (!parseCond(args[3], &cond)) {
          return false;
        }
        instruction.setArg(3, cond);

        if (cond == Cond_Marker) {
          if (args.size() != 5) {
            return false;
          }

          int marker;
          if (!parseInt(args[4], &marker)) {
            return false;
          }
          instruction.setArg(4, marker);
        } else {
          if (args.size() != 4) {
            return false;
          }
        }
      }
      break;
    case InstructionType_Mark:
      {
        if (args.size() != 2) {
          return false;
        }

        int mark;
        if (!parseInt(args[0], &mark)) {
          return false;
        }
        instruction.setArg(0, mark);

        int st;
        if (!parseInt(args[1], &st)) {
          return false;
        }
        instruction.setArg(1, st);
      }
      break;
    case InstructionType_Unmark:
      {
        if (args.size() != 2) {
          return false;
        }

        int mark;
        if (!parseInt(args[0], &mark)) {
          return false;
        }
        instruction.setArg(0, mark);

        int st;
        if (!parseInt(args[1], &st)) {
          return false;
        }
        instruction.setArg(1, st);
      }
      break;
    case InstructionType_PickUp:
      {
        if (args.size() != 2) {
          return false;
        }

        int st1;
        if (!parseInt(args[0], &st1)) {
          return false;
        }
        instruction.setArg(0, st1);

        int st2;
        if (!parseInt(args[1], &st2)) {
          return false;
        }
        instruction.setArg(1, st2);
      }
      break;
    case InstructionType_Drop:
      {
        if (args.size() != 1) {
          return false;
        }

        int st;
        if (!parseInt(args[0], &st)) {
          return false;
        }
        instruction.setArg(0, st);
      }
      break;
    case InstructionType_Turn:
      {
        if (args.size() != 2) {
          return false;
        }

        LR lr;
        if (!parseLR(args[0], &lr)) {
          return false;
        }
        instruction.setArg(0, lr);

        int st;
        if (!parseInt(args[1], &st)) {
          return false;
        }
        instruction.setArg(1, st);
      }
      break;
    case InstructionType_Move:
      {
        if (args.size() != 2) {
          return false;
        }

        int st1;
        if (!parseInt(args[0], &st1)) {
          return false;
        }
        instruction.setArg(0, st1);

        int st2;
        if (!parseInt(args[1], &st2)) {
          return false;
        }
        instruction.setArg(1, st2);
      }
      break;
    case InstructionType_Flip:
      {
        if (args.size() != 3) {
          return false;
        }

        int p;
        if (!parseInt(args[0], &p)) {
          return false;
        }
        instruction.setArg(0, p);

        int st1;
        if (!parseInt(args[1], &st1)) {
          return false;
        }
        instruction.setArg(1, st1);

        int st2;
        if (!parseInt(args[2], &st2)) {
          return false;
        }
        instruction.setArg(2, st2);
      }
      break;
    }

    pBrain->addInstruction(instruction);
  }

  return true;
}

bool formatBrain(const Brain &brain, string *pStr)
{
  int numInstructions = brain.getNumInstructions();
  for (int instructionNum = 0;
       instructionNum < numInstructions;
       instructionNum++) {
    const Instruction &instruction = brain.getInstruction(instructionNum);

    vector<string> args;
    int numArgs = instruction.getNumArgs();
    args.resize(numArgs);

    switch (instruction.getInstructionType()) {
    case InstructionType_Sense:
      {
        pStr->append("Sense");
        if (numArgs != 4 && numArgs != 5) {
          return false;
        }
        if (!formatSenseDir(SenseDir(instruction.getArg(0)), &args[0])) {
          return false;
        }
        if (!formatInt(instruction.getArg(1), &args[1])) {
          return false;
        }
        if (!formatInt(instruction.getArg(2), &args[2])) {
          return false;
        }
        if (!formatCond(Cond(instruction.getArg(3)), &args[3])) {
          return false;
        }
        if (Cond(instruction.getArg(3)) == Cond_Marker) {
          if (numArgs != 5) {
            return false;
          }
          if (!formatInt(instruction.getArg(4), &args[4])) {
            return false;
          }
        } else {
          if (numArgs != 4) {
            return false;
          }
        }
      }
      break;
    case InstructionType_Mark:
      {
        pStr->append("Mark");
        if (numArgs != 2) {
          return false;
        }
        if (!formatInt(instruction.getArg(0), &args[0])) {
          return false;
        }
        if (!formatInt(instruction.getArg(1), &args[1])) {
          return false;
        }
      }
      break;
    case InstructionType_Unmark:
      {
        pStr->append("Unmark");
        if (numArgs != 2) {
          return false;
        }
        if (!formatInt(instruction.getArg(0), &args[0])) {
          return false;
        }
        if (!formatInt(instruction.getArg(1), &args[1])) {
          return false;
        }
      }
      break;
    case InstructionType_PickUp:
      {
        pStr->append("PickUp");
        if (numArgs != 2) {
          return false;
        }
        if (!formatInt(instruction.getArg(0), &args[0])) {
          return false;
        }
        if (!formatInt(instruction.getArg(1), &args[1])) {
          return false;
        }
      }
      break;
    case InstructionType_Drop:
      {
        pStr->append("Drop");
        if (numArgs != 1) {
          return false;
        }
        if (!formatInt(instruction.getArg(0), &args[0])) {
          return false;
        }
      }
      break;
    case InstructionType_Turn:
      {
        pStr->append("Turn");
        if (numArgs != 2) {
          return false;
        }
        if (!formatLR(LR(instruction.getArg(0)), &args[0])) {
          return false;
        }
        if (!formatInt(instruction.getArg(1), &args[1])) {
          return false;
        }
      }
      break;
    case InstructionType_Move:
      {
        pStr->append("Move");
        if (numArgs != 2) {
          return false;
        }
        if (!formatInt(instruction.getArg(0), &args[0])) {
          return false;
        }
        if (!formatInt(instruction.getArg(1), &args[1])) {
          return false;
        }
      }
      break;
    case InstructionType_Flip:
      {
        pStr->append("Flip");
        if (numArgs != 3) {
          return false;
        }
        if (!formatInt(instruction.getArg(0), &args[0])) {
          return false;
        }
        if (!formatInt(instruction.getArg(1), &args[1])) {
          return false;
        }
        if (!formatInt(instruction.getArg(2), &args[2])) {
          return false;
        }
      }
      break;
    }

    for (int i = 0; i < (int)args.size(); i++) {
      pStr->append(" ");
      pStr->append(args[i]);
    }
    pStr->append("\n");
  }

  return true;
}
