#include <cassert>
#include "Sim.hpp"
#include "World.hpp"
#include "Brain.hpp"

namespace
{
  void adjacentCell(int x, int y, int dir, int *pAdjX, int *pAdjY)
  {
    assert(dir >= 0 && dir <= 5);
    switch (dir) {
    case 0:
      *pAdjX = x + 1;
      *pAdjY = y;
      return;
    case 1:
      if (y % 2 == 0) {
        *pAdjX = x;
        *pAdjY = y + 1;
      } else {
        *pAdjX = x + 1;
        *pAdjY = y + 1;
      }
      return;
    case 2:
      if (y % 2 == 0) {
        *pAdjX = x - 1;
        *pAdjY = y + 1;
      } else {
        *pAdjX = x;
        *pAdjY = y + 1;
      }
      return;
    case 3:
      *pAdjX = x - 1;
      *pAdjY = y;
      return;
    case 4:
      if (y % 2 == 0) {
        *pAdjX = x - 1;
        *pAdjY = y - 1;
      } else {
        *pAdjX = x;
        *pAdjY = y - 1;
      }
      return;
    case 5:
      if (y % 2 == 0) {
        *pAdjX = x;
        *pAdjY = y - 1;
      } else {
        *pAdjX = x + 1;
        *pAdjY = y - 1;
      }
      return;
    }
  }

  int turn(LR lr, int dir)
  {
    switch (lr) {
    case LR_Left:
      return (dir + 5) % 6;
    case LR_Right:
      return (dir + 1) % 6;
    }
    assert(lr == LR_Left || lr == LR_Right);
    return dir;
  }

  void sensedCell(int x, int y, int dir, SenseDir senseDir,
                  int *pSenseX, int *pSenseY)
  {
    switch (senseDir) {
    case SenseDir_Here:
      *pSenseX = x; *pSenseY = y; return;
    case SenseDir_Ahead:
      adjacentCell(x, y, dir, pSenseX, pSenseY); return;
    case SenseDir_LeftAhead:
      adjacentCell(x, y, turn(LR_Left, dir), pSenseX, pSenseY); return;
    case SenseDir_RightAhead:
      adjacentCell(x, y, turn(LR_Right, dir), pSenseX, pSenseY); return;
    }
  }

  bool cellMatches(const World &world, int x, int y, Cond cond, int marker,
                   Color color)
  {
    const Cell &cell = world.cell(x, y);

    if (cell.getCellType() == CellType_Rocky) {
      return (cond == Cond_Rock);
    }

    int antId = cell.getAntId();

    switch (cond) {
    case Cond_Friend:
      return antId != -1 &&
        world.ant(antId).getColor() == color;
    case Cond_Foe:
      return antId != -1 &&
        world.ant(antId).getColor() != color;
    case Cond_FriendWithFood:
      return antId != -1 &&
        world.ant(antId).getColor() == color &&
        world.ant(antId).getHasFood();
    case Cond_FoeWithFood:
      return antId != -1 &&
        world.ant(antId).getColor() != color &&
        world.ant(antId).getHasFood();
    case Cond_Food:
      return cell.getFood() > 0;
    case Cond_Rock:
      return false;
    case Cond_Marker:
      return cell.getMarker(color, marker);
    case Cond_FoeMarker:
      return cell.getAnyMarker(otherColor(color));
    case Cond_Home:
      return
        (cell.getCellType() == CellType_RedAnthill && color == Color_Red) ||
        (cell.getCellType() == CellType_BlackAnthill && color == Color_Black);
    case Cond_FoeHome:
      return
        (cell.getCellType() == CellType_RedAnthill && color == Color_Black) ||
        (cell.getCellType() == CellType_BlackAnthill && color == Color_Red);
    }

    assert(cond != cond);
    return false;
  }

  int adjacentAnts(const World &world, int x, int y, Color color)
  {
    int n = 0;
    for (int dir = 0; dir < 6; dir++) {
      int adjX, adjY;
      adjacentCell(x, y, dir, &adjX, &adjY);
      const Cell &cell = world.cell(adjX, adjY);
      int antId = cell.getAntId();
      if (antId != -1) {
        const Ant &ant = world.ant(antId);
        if (ant.getColor() == color) {
          n++;
        }
      }
    }
    return n;
  }

  void checkForSurroundedAntAt(World *pWorld, int x, int y)
  {
    Cell &cell = pWorld->cell(x, y);
    int antId = cell.getAntId();
    if (antId != -1) {
      Ant &ant = pWorld->ant(antId);
      if (adjacentAnts(*pWorld, x, y, otherColor(ant.getColor())) >= 5) {
        cell.setAntId(-1);
        ant.setX(-1);
        ant.setY(-1);
        int delta = 3 + (ant.getHasFood() ? 1 : 0);
        cell.setFood(cell.getFood() + delta);
        pWorld->adjustScore(cell.getCellType(), delta);
      }
    }
  }

  void checkForSurroundingAnts(World *pWorld, int x, int y)
  {
    checkForSurroundedAntAt(pWorld, x, y);
    for (int dir = 0; dir < 6; dir++) {
      int adjX, adjY;
      adjacentCell(x, y, dir, &adjX, &adjY);
      checkForSurroundedAntAt(pWorld, adjX, adjY);
    }
  }
}

void step(World *pWorld, const Brain brains[Color_Count])
{
  for (int antId = 0; antId < pWorld->getNumAnts(); antId++) {
    Ant &ant = pWorld->ant(antId);

    int x = ant.getX();
    int y = ant.getY();
    if (x < 0 || y < 0) { // This means the ant is dead
      continue;
    }

    if (ant.getResting() > 0) {
      ant.setResting(ant.getResting() - 1);
      continue;
    }

    int state = ant.getState();
    Color color = ant.getColor();
    assert(state < brains[color].getNumInstructions());

    int dir = ant.getDir();

    Cell &cell = pWorld->cell(x, y);

    const Instruction &instruction = brains[color].getInstruction(state);
    InstructionType instructionType = instruction.getInstructionType();
    switch (instructionType) {
    case InstructionType_Sense:
      {
        SenseDir senseDir = SenseDir(instruction.getArg(0));
        int st1 = instruction.getArg(1);
        int st2 = instruction.getArg(2);
        Cond cond = Cond(instruction.getArg(3));
        int marker = (cond == Cond_Marker) ? instruction.getArg(4) : 0;

        int senseX, senseY;
        sensedCell(x, y, dir, senseDir, &senseX, &senseY);

        int newState;
        if (cellMatches(*pWorld, senseX, senseY, cond, marker, color)) {
          newState = st1;
        } else {
          newState = st2;
        }

        ant.setState(newState);
      }
      break;
    case InstructionType_Mark:
      {
        int marker = instruction.getArg(0);
        int st = instruction.getArg(1);

        cell.setMarker(color, marker, true);
        ant.setState(st);
      }
      break;
    case InstructionType_Unmark:
      {
        int marker = instruction.getArg(0);
        int st = instruction.getArg(1);

        cell.setMarker(color, marker, false);
        ant.setState(st);
      }
      break;
    case InstructionType_PickUp:
      {
        int st1 = instruction.getArg(0);
        int st2 = instruction.getArg(1);

        if (ant.getHasFood() || cell.getFood() == 0) {
          ant.setState(st2);
        } else {
          cell.setFood(cell.getFood() - 1);
          pWorld->adjustScore(cell.getCellType(), -1);
          ant.setHasFood(true);
          ant.setState(st1);
        }
      }
      break;
    case InstructionType_Drop:
      {
        int st = instruction.getArg(0);

        if (ant.getHasFood()) {
          cell.setFood(cell.getFood() + 1);
          pWorld->adjustScore(cell.getCellType(), 1);
          ant.setHasFood(false);
        }
        ant.setState(st);
      }
      break;
    case InstructionType_Turn:
      {
        LR lr = LR(instruction.getArg(0));
        int st = instruction.getArg(1);

        ant.setDir(turn(lr, ant.getDir()));
        ant.setState(st);
      }
      break;
    case InstructionType_Move:
      {
        int st1 = instruction.getArg(0);
        int st2 = instruction.getArg(1);

        int newX, newY;
        adjacentCell(x, y, ant.getDir(), &newX, &newY);
        Cell &newCell = pWorld->cell(newX, newY);

        if (newCell.getCellType() == CellType_Rocky ||
            newCell.getAntId() != -1) {
          ant.setState(st2);
        } else {
          cell.setAntId(-1);
          newCell.setAntId(antId);
          ant.setX(newX);
          ant.setY(newY);
          ant.setState(st1);
          ant.setResting(14);
          checkForSurroundingAnts(pWorld, newX, newY);
        }
      }
      break;
    case InstructionType_Flip:
      {
        int p = instruction.getArg(0);
        int st1 = instruction.getArg(1);
        int st2 = instruction.getArg(2);

        int r = pWorld->randomInt(p);
        ant.setState(r == 0 ? st1 : st2);
      }
      break;
    }
  }
}
