#include <cassert>
#include <cstdio>
#include "World.hpp"
#include "File.hpp"

Cell::Cell()
: m_cellType(CellType_Clear)
, m_antId(-1)
, m_food(0)
{
  for (int colorInt = Color_First; colorInt <= Color_Last; colorInt++) {
    Color color = Color(colorInt);
    for (int i = 0; i < 6; i++) {
      m_markers[color][i] = false;
    }
  }
}

Cell::~Cell()
{
}

bool Cell::getMarker(Color color, int marker) const
{
  assert(color >= Color_First && color <= Color_Last);
  assert(marker >= 0 && marker <= 5);
  return m_markers[color][marker];
}

void Cell::setMarker(Color color, int marker, bool value)
{
  assert(color >= Color_First && color <= Color_Last);
  assert(marker >= 0 && marker <= 5);
  m_markers[color][marker] = value;
}

bool Cell::getAnyMarker(Color color) const
{
  assert(color >= Color_First && color <= Color_Last);
  for (int i = 0; i < 6; i++) {
    if (m_markers[color][i]) {
      return true;
    }
  }
  return false;
}

Ant::Ant()
: m_color(Color_Red)
, m_state(0)
, m_resting(0)
, m_dir(0)
, m_hasFood(false)
, m_x(-1)
, m_y(-1)
{
}

Ant::~Ant()
{
}

World::World()
: m_width(0)
, m_height(0)
{
  setSeed(12345);

  for (int colorInt = Color_First;
       colorInt <= Color_Last;
       colorInt++) {
    Color color = Color(colorInt);
    m_scores[color] = 0;
  }
}

World::~World()
{
}

void World::setSize(int width, int height)
{
  m_width = width;
  m_height = height;
  m_cells.resize(0);
  m_cells.resize(width*height);
}

int World::getNumAnts() const
{
  return m_ants.size();
}

void World::setNumAnts(int numAnts)
{
  m_ants.resize(numAnts);
}

const Cell &World::cell(int x, int y) const
{
  assert(x >= 0 && x < m_width);
  assert(y >= 0 && y < m_height);
  return m_cells[y * m_width + x];
}

Cell &World::cell(int x, int y)
{
  assert(x >= 0 && x < m_width);
  assert(y >= 0 && y < m_height);
  return m_cells[y * m_width + x];
}

const Ant &World::ant(int antId) const
{
  assert(antId >= 0 && antId < (int)m_ants.size());
  return m_ants[antId];
}

Ant &World::ant(int antId)
{
  assert(antId >= 0 && antId < (int)m_ants.size());
  return m_ants[antId];
}

namespace
{
  void incrementSeed(unsigned int *pSeed)
  {
    *pSeed = *pSeed * 22695477 + 1;
  }
}

void World::setSeed(unsigned int seed)
{
  m_seed = seed;
  for (int i = 0; i < 4; i++) {
    incrementSeed(&m_seed);
  }
}

int World::randomInt(int n)
{
  int result = ((m_seed >> 16) % 16384) % n;
  incrementSeed(&m_seed);
  return result;
}

int World::getScore(Color color) const
{
  assert(color >= Color_First && color <= Color_Last);
  return m_scores[color];
}

void World::setScore(Color color, int score)
{
  assert(color >= Color_First && color <= Color_Last);
  m_scores[color] = score;
}

void World::adjustScore(CellType cellType, int delta)
{
  if (cellType == CellType_RedAnthill) {
    setScore(Color_Red, getScore(Color_Red) + delta);
  } else if (cellType == CellType_BlackAnthill) {
    setScore(Color_Black, getScore(Color_Black) + delta);
  }
}

bool readWorld(const string &fileName, World *pWorld)
{
  vector<unsigned char> buf;
  if (!readFile(fileName, &buf)) {
    return false;
  }

  int pos = 0;
  int sz = buf.size();

  int width = 0;
  int height = 0;

  while (pos < sz && buf[pos] != '\n') {
    if (buf[pos] < '0' || buf[pos] > '9') {
      return false;
    }
    width *= 10;
    width += (buf[pos] - '0');
    pos++;
  }
  if (pos >= sz) {
    return false;
  }
  pos++;

  while (pos < sz && buf[pos] != '\n') {
    if (buf[pos] < '0' || buf[pos] > '9') {
      return false;
    }
    height *= 10;
    height += (buf[pos] - '0');
    pos++;
  }
  if (pos >= sz) {
    return false;
  }
  pos++;

  pWorld->setSize(width, height);
  for (int y = 0; y < height; y++) {
    for (int x = 0; x < width; x++) {
      if (!(x == 0 && y % 2 == 0)) {
        if (pos >= sz) {
          return false;
        }
        if (buf[pos] != ' ') {
          return false;
        }
        pos++;
      }

      if (pos >= sz) {
        return false;
      }
      unsigned char ch = buf[pos];
      pos++;

      Cell &cell = pWorld->cell(x, y);

      if (ch == '#') {
        cell.setCellType(CellType_Rocky);
      } else if (ch == '.') {
        cell.setCellType(CellType_Clear);
      } else if (ch == '+' || ch == '-') {
        int antId = pWorld->getNumAnts();
        pWorld->setNumAnts(antId + 1);
        Ant &ant = pWorld->ant(antId);
        cell.setAntId(antId);
        ant.setX(x);
        ant.setY(y);
        if (ch == '+') {
          cell.setCellType(CellType_RedAnthill);
          ant.setColor(Color_Red);
        } else {
          cell.setCellType(CellType_BlackAnthill);
          ant.setColor(Color_Black);
        }
      } else if (ch >= '1' && ch <= '9') {
        cell.setCellType(CellType_Clear);
        cell.setFood(ch - '0');
      } else {
        return false;
      }
    }

    // Allow extra spaces at the end of the line
    while (pos < sz && buf[pos] == ' ') {
      pos++;
    }

    if (pos >= sz) {
      return false;
    }
    if (buf[pos] != '\n') {
      return false;
    }
    pos++;
  }

  if (pos < sz) {
    return false;
  }

  // Set up ants and stuff here

  return true;
}

void printWorld(const World &world)
{
  for (int y = 0; y < world.getHeight(); y++) {
    for (int x = 0; x < world.getWidth(); x++) {
      if (x == 0 && y % 2 == 1) {
        printf(" ");
      }

      const Cell &cell = world.cell(x, y);
      int antId = cell.getAntId();
      int food = cell.getFood();
      CellType cellType = cell.getCellType();
      if (antId != -1) {
        const Ant &ant = world.ant(antId);
        if (ant.getColor() == Color_Red) {
          printf("$ ");
        } else {
          printf("* ");
        }
      } else if (food > 0) {
        if (food < 10) {
          printf("%c ", '0' + food);
        } else if (food < 36) {
          printf("%c ", 'A' + (food - 10));
        } else if (food < 62) {
          printf("%c ", 'a' + (food - 36));
        } else {
          printf("z ");
        }
      } else if (cellType == CellType_Rocky) {
        printf("# ");
      } else if (cellType == CellType_RedAnthill) {
        printf("+ ");
      } else if (cellType == CellType_BlackAnthill) {
        printf("- ");
      } else {
        printf(" . ");
      }
    }
    printf("\n");
  }
}

namespace
{
  void changeColor(string *pCurColor, const string &newColor)
  {
    if (pCurColor->compare(newColor) != 0) {
      printf("%s", newColor.c_str());
      *pCurColor = newColor;
    }
  }
}

void printWorldAnsi(const World &world)
{
  for (int y = 0; y < world.getHeight(); y++) {
    string curColor;
    for (int x = 0; x < world.getWidth(); x++) {
      if (x == 0 && y % 2 == 1) {
        printf(" ");
      }

      const Cell &cell = world.cell(x, y);
      int antId = cell.getAntId();
      int food = cell.getFood();
      CellType cellType = cell.getCellType();
      if (antId != -1) {
        const Ant &ant = world.ant(antId);
        if (ant.getColor() == Color_Red) {
          changeColor(&curColor, "\x1b[0;1;37;41m");
        } else {
          changeColor(&curColor, "\x1b[0;1;37m");
        }
        if (ant.getHasFood()) {
          printf(" .");
        } else {
          printf("  ");
        }
      } else if (food > 0) {
        changeColor(&curColor, "\x1b[0;1;37;45m");
        printf("%2d", food > 99 ? 99 : food);
      } else if (cellType == CellType_Rocky) {
        changeColor(&curColor, "\x1b[0;44m");
        printf("  ");
      } else if (cellType == CellType_RedAnthill) {
        changeColor(&curColor, "\x1b[0;31;42m");
        printf(" +");
      } else if (cellType == CellType_BlackAnthill) {
        changeColor(&curColor, "\x1b[0;30;42m");
        printf(" -");
      } else {
        changeColor(&curColor, "\x1b[0;42m");
        /*
        if (cell.getAnyMarker(Color_Red) ||
            cell.getAnyMarker(Color_Black)) {
          printf(". ");
        } else {
          printf("  ");
        }
        */
        if (//cell.getMarker(Color_Red, 0) ||
            //cell.getMarker(Color_Red, 1) ||
            //cell.getMarker(Color_Red, 2) ||
            //cell.getMarker(Color_Red, 3) ||
            //cell.getMarker(Color_Red, 4) ||
            //cell.getMarker(Color_Red, 5) ||
            //cell.getMarker(Color_Black, 0) ||
            //cell.getMarker(Color_Black, 1) ||
            //cell.getMarker(Color_Black, 2) ||
            cell.getMarker(Color_Black, 3) ||
            cell.getMarker(Color_Black, 4) ||
            cell.getMarker(Color_Black, 5) ||
            false) {
          printf(". ");
        } else {
          printf("  ");
        }
      }
    }
    printf("\x1b[0m\n");
  }
}

void printWorldInOfficialFormat(const World &world)
{
  printf("%d\n", world.getWidth());
  printf("%d\n", world.getHeight());

  for (int y = 0; y < world.getHeight(); y++) {
    for (int x = 0; x < world.getWidth(); x++) {
      if (x == 0 && y % 2 == 1) {
        printf(" ");
      }

      const Cell &cell = world.cell(x, y);
      int food = cell.getFood();
      CellType cellType = cell.getCellType();
      if (food > 0) {
        printf("%c ", '0' + (food > 9 ? 9 : food));
      } else if (cellType == CellType_Rocky) {
        printf("# ");
      } else if (cellType == CellType_RedAnthill) {
        printf("+ ");
      } else if (cellType == CellType_BlackAnthill) {
        printf("- ");
      } else {
        printf(". ");
      }
    }
    printf("\n");
  }
}

void dumpWorld(const World &world)
{
  for (int y = 0; y < world.getHeight(); y++) {
    for (int x = 0; x < world.getWidth(); x++) {
      printf("cell (%d, %d): ", x, y);

      const Cell &cell = world.cell(x, y);
      CellType cellType = cell.getCellType();
      if (cellType == CellType_Rocky) {
        printf("rock");
      } else {
        if (cell.getFood() > 0) {
          printf("%d food; ", cell.getFood());
        }

        if (cellType == CellType_RedAnthill) {
          printf("red hill; ");
        } else if (cellType == CellType_BlackAnthill) {
          printf("black hill; ");
        }

        for (int colorInt = Color_First;
             colorInt <= Color_Last;
             colorInt++) {
          Color color = Color(colorInt);
          if (cell.getAnyMarker(color)) {
            printf(color == Color_Red ? "red marks: " : "black marks: ");
            for (int marker = 0; marker < 6; marker++) {
              if (cell.getMarker(color, marker)) {
                printf("%d", marker);
              }
            }
            printf("; ");
          }
        }

        int antId = cell.getAntId();
        if (antId != -1) {
          const Ant &ant = world.ant(antId);
          printf(ant.getColor() == Color_Red ? "red" : "black");
          printf(" ant of id %d, dir %d, food %d, state %d, resting %d",
                 antId,
                 ant.getDir(),
                 ant.getHasFood() ? 1 : 0,
                 ant.getState(),
                 ant.getResting());
        }
      }

      printf("\n");
    }
  }
}
