> {-# OPTIONS -Wall -fno-warn-unused-matches #-}

> module AntAsmParser (pProgram) where

> import Char
> import Monad
> import Parsec
> import ParsecToken
> import ParsecLanguage

> import AntAsmInstrs

> pProgram :: Parser [(SourcePos, SourceStmt)]
> pProgram = do pWhitespace
>               stmts <- many pStmt
>               eof
>               return stmts

> pStmt :: Parser (SourcePos, SourceStmt)
> pStmt =
>     do pos <- getPosition
>        stmt <- (fmap Instr pInstr <|> fmap Label pLabel)
>        return (pos, stmt)

Parse ant state machine instruction

> pInstr :: Parser SourceInstr
> pInstr = pInstr' <?> "instruction"
> pInstr' :: Parser SourceInstr
> pInstr' =
>     do pKeyword "sense"
>        dir <- pParamOnDirAndFood pSenseDir
>        nextState <- pStates
>        badState <- pStates
>        cond <- pParamOnDirAndFood pCondition
>        return (Sense dir nextState badState cond)
>     <|>
>     do pKeyword "mark"
>        marker <- pMarkers
>        nextState <- pStates
>        return (Mark marker nextState)
>     <|>
>     do pKeyword "unmark"
>        marker <- pMarkers
>        nextState <- pStates
>        return (Unmark marker nextState)
>     <|>
>     do pKeyword "pickup"
>        nextState <- pStates
>        badState <- pStates
>        return (PickUp nextState badState)
>     <|>
>     do pKeyword "drop"
>        nextState <- pStates
>        return (Drop nextState)
>     <|>
>     do pKeyword "turn"
>        dir <- pParamOnDirAndFood pTurnDir
>        nextState <- pStates
>        return (Turn dir nextState)
>     <|>
>     do pKeyword "move"
>        nextState <- pStates
>        badState <- pStates
>        return (Move nextState badState)
>     <|>
>     do pKeyword "flip"
>        chance <- pParamOnDirAndFood pNatural
>        trueState <- pStates
>        falseState <- pStates
>        return (Flip chance trueState falseState)

Parse label declaration

> pLabel :: Parser String
> pLabel =
>     do name <- pIdentifier <?> "label"
>        pColon
>        return name

Parse a list of things possibly parametrized on direction we're facing

> pParamOnDirection :: Parser item -> Parser [item]
> pParamOnDirection pItem =
>     do items <- pInAngles (many1 pItem)
>        when (length items /= 6)
>           (fail ("parametrizing on direction requires 6 elements, not " ++
>                  show (length items)))
>        return items
>     <|> fmap (:[]) pItem

Parse a list of things possibly parametrized on whether we're carrying food

> pParamOnFood :: Parser item -> Parser [item]
> pParamOnFood pItem =
>     do items <- pInBrackets (many1 pItem)
>        when (length items /= 2)
>           (fail ("parametrizing on foodp requires 2 elements, not " ++
>                  show (length items)))
>        return items
>     <|> fmap (:[]) pItem

Parse a list of lists of things, possibly parametrized on direction and foodp

> pParamOnDirAndFood :: Parser item -> Parser [[item]]
> pParamOnDirAndFood = pParamOnDirection . pParamOnFood

Parse a list of states or one state

> pStates :: Parser [[StateRef]]
> pStates = pParamOnDirAndFood pState <?> "states"

Parse label reference or absolute address

> pState :: Parser StateRef
> pState =
>     (pKeywordInto "next" Next
>      <|> pKeywordInto "afternext" AfterNext
>      <|> fmap Absolute pNatural
>      <|> fmap LabelRef pIdentifier) <?> "state"

Parse turn direction

> pTurnDir :: Parser TurnDir
> pTurnDir = pKeywordTable buildKeywordTable <?> "turn direction"

Parse sense direction

> pSenseDir :: Parser SenseDir
> pSenseDir = pKeywordTable buildKeywordTable <?> "sense direction"

Parse sense condition

> pCondition :: Parser (Condition [[Marker]])
> pCondition =
>     (do { pKeyword "marker"; marker <- pMarkers; return (Marker marker) }
>     <|> pKeywordTable condTable) <?> "sense condition"
>     where condTable = [("friend", Friend), ("foe", Foe),
>                        ("friendwithfood", FriendWithFood),
>                        ("foewithfood", FoeWithFood),
>                        ("food", Food), ("rock", Rock),
>                        ("foemarker", FoeMarker), ("home", Home),
>                        ("foehome", FoeHome)]

> pMarkers :: Parser [[Marker]]
> pMarkers = pParamOnDirAndFood pMarker

> pMarker :: Parser Marker
> pMarker = pNatural -- XXX limit range

Lexical conventions for ant language

> antLangDef :: LanguageDef parsestate
> antLangDef =
>     LanguageDef { commentStart = "/*",
>                   commentEnd = "*/",
>                   commentLine = ";",
>                   nestedComments = True,
>                   identStart = letter <|> char '_',
>                   identLetter = alphaNum <|> char '_',
>                   opStart = oneOf "<>",
>                   opLetter = oneOf "<>",
>                   reservedNames = [-- instructions
>                                    "sense", "mark", "unmark",
>                                    "pickup", "drop", "turn",
>                                    "move", "flip",
>                                    -- turn directions
>                                    "left", "right",
>                                    -- sense directions
>                                    "here", "ahead", "leftahead",
>                                    "rightahead",
>                                    -- conditions
>                                    "friend", "foe",
>                                    "friendwithfood", "foewithfood",
>                                    "food", "rock", "marker",
>                                    "foemarker", "home", "foehome",
>                                    -- special states
>                                    "next", "afternext"],
>                            reservedOpNames = ["<", ">"],
>                            caseSensitive = False }

Token parser for ant language

> antTokenizer :: TokenParser parsestate
> antTokenizer = makeTokenParser antLangDef

Parse a given keyword

> pKeyword :: String -> Parser ()
> pKeyword = reserved antTokenizer

Parse a given keyword and return a given result

> pKeywordInto :: String -> item -> Parser item
> pKeywordInto keyword result = pKeyword keyword >> return result

Parse a keyword table

> pKeywordTable :: [(String, item)] -> Parser item
> pKeywordTable [] = fail "no keywords in table"
> pKeywordTable table = choice [pKeywordInto kw result | (kw, result) <- table]

Parse a reserved operator

> pOp :: String -> Parser ()
> pOp = reservedOp antTokenizer

> pNatural :: Parser Int
> pNatural = fmap fromInteger (natural antTokenizer) <?> "number"

> pIdentifier :: Parser String
> pIdentifier = identifier antTokenizer <?> "identifier"

> pColon :: Parser String
> pColon = colon antTokenizer

> pWhitespace :: Parser ()
> pWhitespace = whiteSpace antTokenizer

Build a keyword table from a data type, provided it can be enumerated and shown

> buildKeywordTable :: (Enum item, Bounded item, Show item) => [(String, item)]
> buildKeywordTable =
>     [(map toLower (show dir), dir) | dir <- [minBound .. maxBound]]

> pInAngles :: Parser result -> Parser result
> pInAngles parser =
>     do pOp "<"
>        result <- parser
>        pOp ">"
>        return result

> pInBrackets :: Parser result -> Parser result
> pInBrackets = brackets antTokenizer
